#include "header.h"

// ForceDirectDrawEmulation by passing DDCREATE_EMULATIONONLY flag to IDDraw->Initialize()
// ( see DirectDrawCreate[Ex] or dd(1...7)->Initialize() )


// ColorFix / Stolen Palette fix sets flag 0x800 on a internal ddraw struct member called dwAppHackFlags
//  ( see DirectDrawCreate[Ex] or dd(1...7)->Initialize() )
// the only other fla I know is:
// if ( flag & 0x10 ) do not disable the screen saver for exclusive mode 
// else disable by calling SystemParametersInfo( SPI_SCREENSAVERRUNNING, 1 ...) 
// note: SystemParametersInfo( SPI_SCREENSAVERRUNNING, ...) disables alt+tab and ctrl+alt+del on win9x

/*
todo:
1a. HEL and "Windows 7 compatible applications must not lock the primary video buffer to render." doing so results in an error...
	- this is a big deal because HEL is now preferred over HAL... by me.

1b. "Outside of the lock/unlock pair, the surface structure is not guaranteed to reside at the same location, nor even to exist." ( Since Windows 2000 )
	- programs shouldn't be doing this anyways but doesn't mean that they don't.... ( *cough* "Axis & Allies: Iron Blitz" *cough* )
	- this behavior breaks Wine as well AFAIK

2. "Applications manifested for Windows 7 are prevented from performing Blt's to the primary Desktop video buffer without a clipping window."
	- who cares; don't manifested your directdraw app for win7... however the DWM Composition stays on if you do
	- BltFast doesn't work with a clipper, so redirect it thru Blt
	- grab the hWnd from SetCooperativeLevel
	
*/

// HEL doesn't support (or fake) vsync 
// this is a problem because some animation is frame based and can run too fast
// we can get the real verticle blank from D3DKMT but the screen still tears in HEL though I don't know why???
// todo: what is GetVerticalBlankStatus returning in HEL?
// todo: DDOVER_AUTOFLIP ??
HRESULT WaitForVerticalBlankProc( WRAP* wrap, DWORD dwFlags, HANDLE hEvent )
{ 
	HRESULT hResult;
	static DWORD dwPrevFrameTickCount = 0;
	
	// use real vsync if possible
	hResult = wrap->dd1->WaitForVerticalBlank( dwFlags, hEvent ); 

	// else fake it
	if( FAILED( hResult ) && Config.FakeVsync ){ 
		// 16ms is about the time between frames when running 60fps
		DWORD TickCount = GetTickCount();
		LONG Diff = 16 - (TickCount - dwPrevFrameTickCount);
		if(Diff > 0){
			Sleep(Diff);
			TickCount += Diff;
		}
		dwPrevFrameTickCount = TickCount;
		if( hEvent ) SetEvent( hEvent ); //todo: PulseEvent instead ???
		hResult = DD_OK;
	}
	return hResult;
}  

/*
DPI virtualization causes:
Text Clipping, Blurring, or Inconsistent font sizes.
"Rendering of full-screen DX applications partially off screen" - Mircosoft
...drawing(writting) to someplace that doesn't exist may cause crashes... 
if your going to Disable DPI Scaling then do it as soon as possible
*/
void DisableHighDPIScaling()
{
	// use GetProcAddress because SetProcessDPIAware exists only on win6+
	// and "High" dpi scaling only exits on win6+?
	HMODULE hUser32 = GetModuleHandle( "user32.dll" );
	typedef BOOL (__stdcall* SetProcessDPIAwareFunc)();
	if( hUser32 )
	{
		SetProcessDPIAwareFunc setDPIAware = (SetProcessDPIAwareFunc)GetProcAddress(hUser32, "SetProcessDPIAware");
		if (setDPIAware) setDPIAware();
	}
}


// HEL and some dual graphic cards setups discard the front buffer instead of returning it to the back buffer
// so we'll do that by hand... 
struct FullFlipContext
{
	LPDIRECTDRAWSURFACE lpdds;
	DWORD dwWidth;
	DWORD dwHeight;
	DWORD dwDescSize; 
};
HRESULT __stdcall EnumFlipSurfCallback( LPDIRECTDRAWSURFACE lpdds, LPDDSURFACEDESC lpddsd, LPVOID lpContext)
{ 
	((FullFlipContext*)lpContext)->lpdds      = lpdds; 
	((FullFlipContext*)lpContext)->dwWidth    = lpddsd->dwWidth;
	((FullFlipContext*)lpContext)->dwHeight   = lpddsd->dwHeight;
	((FullFlipContext*)lpContext)->dwDescSize = lpddsd->dwSize;

	return DDENUMRET_OK; // enum until the end
}
HRESULT FlipProc( WRAP* wrap, LPDIRECTDRAWSURFACE lpDDSurfaceTargetOverride, DWORD dwFlags )
{
	lpDDSurfaceTargetOverride = GetInterfacePtr( lpDDSurfaceTargetOverride );
	IDirectDraw* ddraw = GetInterfacePtr( ((IDirectDraw*)(wrap->lparam)) );

	if( Config.FullFlip )
	{
		if( ddraw ) 
		{
			FullFlipContext context = { 0, 0, 0, 0 };
			if( SUCCEEDED( wrap->dds1->EnumAttachedSurfaces( &context, &EnumFlipSurfCallback ) ) )
			{
				LPDIRECTDRAWSURFACE lpDDSurfaceWork;
				DDSURFACEDESC2 ddsd;
				HRESULT hResult;

				memset(&ddsd, 0, sizeof(ddsd));
				ddsd.dwSize = context.dwDescSize;
				ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
				ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
				ddsd.dwWidth = context.dwWidth;
				ddsd.dwHeight = context.dwHeight;
				hResult = ddraw->CreateSurface((LPDDSURFACEDESC)&ddsd, &lpDDSurfaceWork, NULL);
				if( FAILED( hResult ) ) // DDERR_OUTOFVIDEOMEMORY maybe
				{
					ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
					hResult = ddraw->CreateSurface((LPDDSURFACEDESC)&ddsd, &lpDDSurfaceWork, NULL);
				}
				if( SUCCEEDED( hResult ))
				{	
					// blt the front buffer to the temp surface
					// flip()
					// blt the temp surface to the back buffer OR lpDDSurfaceTargetOverride
					// note: Blt is used because BltFast doesn't work with a clipper
					hResult = lpDDSurfaceWork->Blt( NULL, wrap->dds1, NULL, DDBLT_WAIT, NULL );
					if( SUCCEEDED( hResult ) )
					{
						if( Config.FakeVsync && (!( dwFlags & DDFLIP_NOVSYNC)) )
						{
							WaitForVerticalBlankProc( (WRAP*)(wrap->lparam), DDWAITVB_BLOCKBEGIN, 0 );
							dwFlags |= DDFLIP_NOVSYNC;
						} 
						dwFlags &= ~DDFLIP_DONOTWAIT;
						dwFlags |= DDFLIP_WAIT;

						if( SUCCEEDED( wrap->dds1->Flip( lpDDSurfaceTargetOverride, dwFlags ) ) )
						{
							while( DDERR_WASSTILLDRAWING == wrap->dds1->GetFlipStatus( DDGFS_ISFLIPDONE ) )
							{ // needed? because of DDFLIP_NOVSYNC 
								; //spin 
							} 
	
							if( lpDDSurfaceTargetOverride == NULL ) lpDDSurfaceTargetOverride = context.lpdds;	
							hResult = lpDDSurfaceTargetOverride->Blt( NULL, lpDDSurfaceWork, NULL, DDBLT_WAIT, NULL );
						}
					}
					lpDDSurfaceWork->Release();
					return hResult;
				}
			}
		}
	}

	if( Config.FakeVsync && (!( dwFlags & DDFLIP_NOVSYNC)) )
	{
		WaitForVerticalBlankProc( (WRAP*)(wrap->lparam), DDWAITVB_BLOCKBEGIN, 0 );
		dwFlags |= DDFLIP_NOVSYNC;
	} 
	return wrap->dds1->Flip( lpDDSurfaceTargetOverride, dwFlags );
}


// "Note that the DDBLTFX_NOTEARING, DDBLTFX_MIRRORLEFTRIGHT, and DDBLTFX_MIRRORUPDOWN flags are unsupported 
// on Windows 2000 and later and are never passed to the driver." - http://msdn.microsoft.com/en-us/library/windows/hardware/ff550474
// so here we do those by hand...
// todo: is RopUsesPattern correct with rops or ddrops?
bool RopUsesPattern ( DWORD dwRop ) { return ( (((dwRop & 0x00F00000) >> 4) != (dwRop & 0x000F0000)) || (((dwRop & 0xF0000000) >> 4) != (dwRop & 0x0F000000)) ); }
HRESULT BltProc( WRAP* wrap, LPRECT lpDestRect, LPDIRECTDRAWSURFACE lpDDSrcSurface, LPRECT lpSrcRect, DWORD dwFlags, LPDDBLTFX lpDDBltFx )
{
	DDBLTFX fx;
	IDirectDrawSurface* lpDDWorkSurface = NULL;
	IDirectDraw* ddraw = GetInterfacePtr( ((IDirectDraw*)(wrap->lparam)) );

	lpDDSrcSurface = GetInterfacePtr( lpDDSrcSurface ); // unwrap
	if( lpDDBltFx && ( sizeof(DDBLTFX) == lpDDBltFx->dwSize ) )
	{
		// determine which surface pointers should be valid and then unwrap them
		bool UsesPattern = false;
		if( dwFlags & DDBLT_ROP ) UsesPattern = RopUsesPattern( lpDDBltFx->dwROP );		
		if( dwFlags & DDBLT_DDROPS ) UsesPattern |= RopUsesPattern( lpDDBltFx->dwDDROP ); // roll the dice (probably DDBLT_DDROPS will never happen)
		if( UsesPattern || ( dwFlags & ( DDBLT_ALPHADESTSURFACEOVERRIDE | DDBLT_ALPHASRCSURFACEOVERRIDE | DDBLT_ZBUFFERDESTOVERRIDE | DDBLT_ZBUFFERSRCOVERRIDE )) )
		{
			memcpy( &fx, lpDDBltFx, sizeof(fx) );
			if( dwFlags & DDBLT_ALPHADESTSURFACEOVERRIDE ) fx.lpDDSAlphaDest   = GetInterfacePtr( fx.lpDDSAlphaDest   );
			if( dwFlags & DDBLT_ALPHASRCSURFACEOVERRIDE  ) fx.lpDDSAlphaSrc    = GetInterfacePtr( fx.lpDDSAlphaSrc    );
			if( dwFlags & DDBLT_ZBUFFERDESTOVERRIDE      ) fx.lpDDSZBufferDest = GetInterfacePtr( fx.lpDDSZBufferDest );
			if( dwFlags & DDBLT_ZBUFFERSRCOVERRIDE       ) fx.lpDDSZBufferSrc  = GetInterfacePtr( fx.lpDDSZBufferSrc  );
			if( UsesPattern )
			{
				if( ! IsBadReadPtr( fx.lpDDSPattern, 8 ) ) fx.lpDDSPattern = GetInterfacePtr( fx.lpDDSPattern );
			}
			lpDDBltFx = &fx;
		}
		
		if( Config.BltMirror && lpDDSrcSurface && (lpDDBltFx->dwDDFX & (DDBLTFX_MIRRORLEFTRIGHT | DDBLTFX_MIRRORUPDOWN))  ) 
		{
			if( ddraw )
			{
				// if lpSrcRect is NULL then we need to find the width and height of src (this can be done with GDI32.GetDeciceCaps)
				// but we also need to known what size DDSURFACEDESC_X struct to use so we can create a temp surface
				// so we'll determine the size of DDSURFACEDESC_X by querying for the height and width of the src surface 
				// of course we could just pass dwDescSize as an arg to BltProc...
				DDSURFACEDESC2 ddsd;
				DWORD dwDescSize = 0;
				ddsd.dwSize = sizeof( DDSURFACEDESC );
				if( SUCCEEDED( lpDDSrcSurface->GetSurfaceDesc( (LPDDSURFACEDESC)&ddsd ) )) dwDescSize = sizeof( DDSURFACEDESC );
				else{
					ddsd.dwSize = sizeof( DDSURFACEDESC2 ); // try it bigger (v4 and v7)
					if( SUCCEEDED( lpDDSrcSurface->GetSurfaceDesc( (LPDDSURFACEDESC)&ddsd ) )) dwDescSize = sizeof( DDSURFACEDESC2 );
				}
				if( dwDescSize ) 
				{
					DWORD dwX;
					DWORD dwY;
					DWORD dwWidth;
					DWORD dwHeight;
					if( lpSrcRect == NULL )
					{
						dwX = 0;
						dwY = 0;
						dwWidth  = ddsd.dwWidth;					
						dwHeight = ddsd.dwHeight;		
					} else {
						dwX = lpSrcRect->left;
						dwY = lpSrcRect->top;
						dwWidth  = lpSrcRect->right - lpSrcRect->left;
						dwHeight = lpSrcRect->bottom - lpSrcRect->top;
					}

					// create temp surface
					memset( &ddsd, 0, sizeof(ddsd) );
					ddsd.dwSize = dwDescSize;
					ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
					ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
					ddsd.dwWidth = dwWidth;
					ddsd.dwHeight = dwHeight;
					if( SUCCEEDED( ddraw->CreateSurface( (LPDDSURFACEDESC)&ddsd, &lpDDWorkSurface, NULL ) ) )
					{
						if( dwFlags & DDBLT_KEYSRC )
						{
							DDCOLORKEY ddck;
							if( SUCCEEDED( lpDDSrcSurface->GetColorKey( DDCKEY_SRCBLT, &ddck ) ) )
							{
								lpDDWorkSurface->SetColorKey( DDCKEY_SRCBLT, &ddck );
							}
						}

						// mirror src surface onto temp surface
						HDC hdcWork; 
						if( SUCCEEDED( lpDDWorkSurface->GetDC( &hdcWork ) ) )
						{
							HDC hdcSrc;
							if( SUCCEEDED( lpDDSrcSurface->GetDC( &hdcSrc ) ) )
							{	
								//	if( RC_STRETCHBLT & GetDeviceCaps( hdcSrc, RASTERCAPS ) )
								if( StretchBlt( 
									hdcWork, 0, 0, ddsd.dwWidth, ddsd.dwHeight,
									hdcSrc, 
									( lpDDBltFx->dwDDFX & DDBLTFX_MIRRORLEFTRIGHT ) ? (dwX + (dwWidth - 1)) : dwX,
									( lpDDBltFx->dwDDFX & DDBLTFX_MIRRORUPDOWN ) ? (dwY + (dwHeight - 1)) : dwY,
									( lpDDBltFx->dwDDFX & DDBLTFX_MIRRORLEFTRIGHT ) ? (~(dwWidth - 1)) : dwWidth,
									( lpDDBltFx->dwDDFX & DDBLTFX_MIRRORUPDOWN ) ? (~(dwHeight - 1)) : dwHeight,
									SRCCOPY ) )
								{
									lpDDBltFx->dwDDFX &= (~(DDBLTFX_MIRRORLEFTRIGHT | DDBLTFX_MIRRORUPDOWN));
									lpSrcRect = NULL;
									lpDDSrcSurface->ReleaseDC( hdcSrc );
									lpDDSrcSurface = lpDDWorkSurface;
								}
								else lpDDSrcSurface->ReleaseDC( hdcSrc );	
							}
							lpDDWorkSurface->ReleaseDC( hdcWork );	
						}		
					}
				}
			}
		} // end Config.BltMirror
		
		if( Config.BltMirror || Config.FakeVsync )
		{
			if( lpDDBltFx->dwDDFX & DDBLTFX_NOTEARING ){
				WaitForVerticalBlankProc( (WRAP*)(wrap->lparam), DDWAITVB_BLOCKBEGIN, 0 );
				lpDDBltFx->dwDDFX ^= DDBLTFX_NOTEARING;
			}
		}
	}

	// blt it
	HRESULT hResult = wrap->dds1->Blt( lpDestRect, lpDDSrcSurface, lpSrcRect, dwFlags, lpDDBltFx ); 
	if(	lpDDWorkSurface ) lpDDWorkSurface->Release();
	return hResult;
}